import os

from mlsriracha.interfaces.predict import PredictInterface

class AzureMlPredict(PredictInterface):

    def __init__(self):
        print('Selected Azure ML profile')

    def model_artifact(self, filename: str):
        """
        The path to model artifacts.

        Loads from the environmental variable set by mlctl where 
        to retrieve the artifact from Azure ML's mount points. 

        Arguments:
            filename (str): The name of the file which will be written back to S3

        Returns:
            path (str): The absolute path to the model output directory
        """
        model_uri = os.environ.get('AZUREML_MODEL_DIR')
        return os.path.join(model_uri, filename)

    def endpoint_metadata(self):
        return {}


import os
from pathlib import Path
from google.cloud import storage
from google.cloud.storage.blob import Blob

from mlsriracha.interfaces.predict import PredictInterface

class GcpVertexPredict(PredictInterface):

    def __init__(self):
        print('Selected GCP Vertex profile')

        def getBucketNameFrom(gs_uri: str):
            bucket_start = -1
            for i in range(0, 2):
                bucket_start = gs_uri.find('/', bucket_start + 1)

            bucket_end = gs_uri.find('/', bucket_start + 1)
                
            # Printing nth occurrence
            # print ("Nth occurrence is at", val)

            bucket = gs_uri[bucket_start + 1: bucket_end]
            

            prefix = gs_uri[bucket_end + 1: len(gs_uri)]

            # chop off '/' at the end
            if prefix[len(prefix)-1: len(prefix)] == '/':
                prefix = prefix[0: len(prefix) - 1] 
            return bucket, prefix

        # create path for all model data
        Path('/opt/ml/model/').mkdir(parents=True, exist_ok=True)

        storage_client = storage.Client()

        gs_uri = os.getenv('AIP_STORAGE_URI')
        bucket, prefix=getBucketNameFrom(gs_uri)
        # chop off * for wildcard
        prefix = prefix.replace('*', '')
        print(f'bucket_name={bucket}')
        print(f'prefix={prefix}')
        # Note: Client.list_blobs requires at least package version 1.17.0.
        blobs = storage_client.list_blobs(bucket, prefix=prefix)

        # copy all
        for blob in blobs:
            print('Blob: {blob.name}')
            destination_uri = f'/opt/ml/model/{blob.name}' 
            blob.download_to_filename(destination_uri)

    def model_artifact(self, filename: str):
        """
        The path to model artifacts.

        Loads from the environmental variable that mlctl passes to Azure ML

        Arguments:
            filename (str): The name of the file which will be written back to S3

        Returns:
            path (str): The absolute path to the model output directory
        """
        return os.path.join(os.sep, 'opt', 'ml', 'model', filename)

    def endpoint_metadata(self):
        return {
            'container_port': os.getenv('AIP_HTTP_PORT'),
            'model_id': os.getenv('AIP_DEPLOYED_MODEL_ID')
        }


import pandas as pd
import os
import glob
from pathlib import Path

from mlsriracha.interfaces.train import TrainInterface

class AwsSageMakerTrain(TrainInterface):

    def __init__(self):
        print('Selected AWS SageMaker ML profile')
        Path('/opt/ml/model').mkdir(parents=True, exist_ok=True)

    def input_as_dataframe(self, channel='training'):
        """
        The function returns a panda dataframe for the input channel artifacts.

        AWS SageMaker passes all values in FileMode from the S3 bucket into the 
        train, test, and validation "channels" when starting your container.
        This function takes in the channel and merges all CSV files per channel
        into a dataframe for reading.

        Arguments:
            channel (str): The name of the channel which contains the given filename

        Returns:
            path (str): The absolute path to the specified channel file
        """

        data_directories = {
            'training',
            'validation',
            'testing'
        }

        if channel in data_directories:

            csv_files = glob.glob(os.path.join(f'/opt/ml/input/data/{channel}/*.csv'))
            print(f'Files in {channel} directory: {csv_files}')
            # loop over the list of csv files
            fileBytes = []
            for f in csv_files:
        
                # read the csv file
                df = pd.read_csv(f)
                fileBytes.append(df)
            frame = pd.concat(fileBytes, axis=0, ignore_index=True)     
            return frame 
        else:
            raise ValueError('Incorrect data channel type. Options are training, validation, and testing.')
            return null
    
    def log_artifact(self, filename=''):
        """
        The path to the output artifacts.

        Your algorithm should write all final model artifacts to this directory.
        Amazon SageMaker copies this data as a single object in compressed tar
        format to the S3 location that you specified in the CreateTrainingJob
        request. If multiple containers in a single training job write to this
        directory they should ensure no file/directory names clash. Amazon SageMaker
        aggregates the result in a tar file and uploads to S3.

        Arguments:
            filename (str): The name of the file which will be written back to S3

        Returns:
            path (str): The absolute path to the model output directory
        """
        return os.path.join(os.sep, 'opt', 'ml', 'model', filename)

    def finish(self):
        return True


import os

from mlsriracha.interfaces.predict import PredictInterface

class AwsSageMakerPredict(PredictInterface):

    def __init__(self):
        print('Selected AWS SageMaker profile')

    def model_artifact(self, filename: str):
        """
        The path to model artifacts.

        Your algorithm should read all the model artifacts from this directory.
        Amazon SageMaker copies this data as a single object in a compressed tar
        format from the S3 location that you specified in the CreateModel config
        request. If multiple containers in a single training job write to this
        directory they should ensure no file/directory names clash. Amazon SageMaker
        mounts the S3 for the user.

        Arguments:
            filename (str): The name of the file which will be written back to S3

        Returns:
            path (str): The absolute path to the model output directory
        """
        return os.path.join(os.sep, 'opt', 'ml', 'model', filename)

    def endpoint_metadata(self):
        return {}



class PredictInterface:
    def init(self, provider: str):
        """Load the model."""
        pass

    def model_artifact(self):
        """Get the file directory and model path"""
        pass

    def endpoint_metadata():
        pass


import os
from pathlib import Path
import pandas as pd

from google.cloud import storage
from google.cloud.storage.blob import Blob
from io import StringIO

from mlsriracha.interfaces.train import TrainInterface

class GcpVertexTrain(TrainInterface):

    def __init__(self):
        """
        Make folder paths for data, models, checkpoints
        """
        print('Selected GCP Vertex profile')
        Path('/opt/ml/model/').mkdir(parents=True, exist_ok=True)

    def input_as_dataframe(self, channel='training'):
        """
        The function returns a dataframe for the input channel artifacts.

        GCP Vertex allows you to create a dataset that will be split into 
        train, test, and validation "channels" when loaded into your training run.
        This function handles merging the file structure that Vertex 
        exposes to the training container.

        Arguments:
            channel (str): The name of the channel which contains the given filename

        Returns:
            path (str): The absolute path to the specified channel file
        """

        def get_bucket_name_from(gs_uri):
            bucket_start = -1
            for i in range(0, 2):
                bucket_start = gs_uri.find('/', bucket_start + 1)

            bucket_end = gs_uri.find('/', bucket_start + 1)
                
            # Printing nth occurrence
            # print ("Nth occurrence is at", val)

            bucket = gs_uri[bucket_start + 1: bucket_end]
            

            prefix = gs_uri[bucket_end + 1: len(gs_uri)]

            # chop off '/' at the end
            if prefix[len(prefix)-1: len(prefix)] == '/':
                prefix = prefix[0: len(prefix) - 1] 
            return bucket, prefix

        data_directories = {
            'training': "AIP_TRAINING_DATA_URI",
            'validation': "AIP_VALIDATION_DATA_URI",
            'testing': "AIP_TEST_DATA_URI"
        }

        if channel in data_directories:
            print(f'Retrieving {channel} directory')
            gs_uri = os.getenv(data_directories[channel])
            storage_client = storage.Client()

            bucket, prefix=get_bucket_name_from(gs_uri)
            # chop off * for wildcard
            prefix = prefix.replace('*', '')
            print(f'bucket_name={bucket}')
            print(f'prefix={prefix}')
            # Note: Client.list_blobs requires at least package version 1.17.0.
            blobs = storage_client.list_blobs(bucket, prefix=prefix)
            # print(f'blobs: {blobs}')
            fileBytes = []
            for blob in blobs:
                print(f'blob.name: {blob.name}')
                # assumes CSV files
                s=str(blob.download_as_bytes(),'utf-8')
                data = StringIO(s) 
                df=pd.read_csv(data)
                fileBytes.append(df)

            frame = pd.concat(fileBytes, axis=0, ignore_index=True)     
            return frame
        else:
            raise ValueError('Incorrect data channel type. Options are training, validation, and testing.')
            return null
    
    def log_artifact(self, filename=''):
        """
        The path to the output artifacts.

        Your algorithm should write all final model artifacts to this directory.
        Sriracha will upload copies to the GCP provided GS uri for the specific run.

        Arguments:
            filename (str): The name of the file which will be written back to S3

        Returns:
            path (str): The absolute path to the model output directory
        """
        return os.path.join(os.sep, 'opt', 'ml', 'model', filename)

    def finish(self):
        """
        Copies files in local model directory to google storage
        """
        storage_client = storage.Client()
        bucket_uri = os.getenv('AIP_MODEL_DIR')
        bucket_name, prefix = get_bucket_name_from(bucket_uri)
        # blob = bucket.blob('model.pkl')
        # artifact = f'{bucket_uri}/model.pkl'

        bucket = storage_client.bucket(bucket_name)
        for filename in os.listdir('/opt/ml/model/'):
            print(filename)
            blob = bucket.blob(prefix + '/' + filename)
            blob.upload_from_filename('/opt/ml/model/' + filename)
        
        return True



import pandas as pd
class TrainInterface:
    def init(self):
        """Load the model."""
        pass

    def input_as_dataframe(self, channel:str):
        """Return the input data for that channel as a Panda dataframe """
        pass

    def log_artifact(self, filename: str):
        """Returns the directory that the training job should output artifacts to"""
        pass

    def finish(self):
        """Upload all the artifacts saved to the URI provided by the provider"""
        pass


import os
import pandas as pd
from pathlib import Path

from mlsriracha.interfaces.train import TrainInterface

class AzureMlTrain(TrainInterface):

    def __init__(self):
        print('Using Azure ML Sriracha profile')
        Path('./outputs/model').mkdir(parents=True, exist_ok=True)

    def input_as_dataframe(self, channel='training'):
        """
        The path to input artifacts.

        In Azure, mlctl passes environment variables that map to the 
        train.yaml inputs that the user defines.

        Arguments:
            channel (str): The name of the channel which contains the given filename
            filename (str): The name of the file within a specific channel

        Returns:
            path (str): The absolute path to the specified channel file
        """

        # channel --> environmental variable names
        data_directories = {
            'training': "training-data",
            'validation': "validation-data",
            'testing': "testing-data"
        }

        if channel in data_directories.keys():
            azure_mount_file = os.environ.get(data_directories[channel])
            print(f'azure_mount_file={azure_mount_file}')
            data = pd.read_csv(azure_mount_file)
            return data

        else:
            print('Incorrect data channel type. Options are training, validation, and testing.')
            return null
    
    def log_artifact(self, filename=''):
        """
        The path to the output artifacts.

        Your algorithm should write all final model artifacts to this directory.
        Azure ML copies this data as a folder into the console as a Run output.

        Arguments:
            filename (str): The name of the file which will be written back to S3

        Returns:
            path (str): The absolute path to the model output directory
        """
        cwd = os.getcwd()
        return os.path.join(cwd, 'outputs', 'model', filename)

    def finish(self):
        return True


from unittest import mock
import unittest

from mlsriracha.plugins.azureml.train import AzureMlTrain
from mlsriracha.plugins.gcpvertex.train import GcpVertexTrain
from mlsriracha.plugins.awssagemaker.train import AwsSageMakerTrain


class TestTrainingAdapter(unittest.TestCase):

    # @mock.patch('mlsriracha.plugins.awssagemaker.train.Path.mkdir', return_value=[Tru])
    def test_awssagemaker_plugin(self):
        with mock.patch('mlsriracha.plugins.awssagemaker.train.Path.mkdir', autospec=True) as mock_mkdir:
            mock_mkdir.return_value = True
            plugin = AwsSageMakerTrain()
            
        response = plugin.finish()
        self.assertTrue(response)
    
    
    def test_awssagemaker_data(self):
         with mock.patch('mlsriracha.plugins.awssagemaker.train.Path.mkdir', autospec=True) as mock_mkdir:
            with mock.patch('mlsriracha.plugins.awssagemaker.train.glob.glob', autospec=True) as mock_glob:
                mock_mkdir.return_value = True
                mock_glob.return_value = ['./static/data/data.csv']
                plugin = AwsSageMakerTrain()
               
                response = plugin.input_as_dataframe(channel='training')
                # print(response.size)
                self.assertTrue(response.size == 15)

    def test_awssagemaker_folder(self):
         with mock.patch('mlsriracha.plugins.awssagemaker.train.Path.mkdir', autospec=True) as mock_mkdir:
                mock_mkdir.return_value = True
                plugin = AwsSageMakerTrain()
               
                response = plugin.log_artifact('model.pkl')
                print(response)
                self.assertTrue(response == '/opt/ml/model/model.pkl')

    def test_azureml_plugin(self):
        with mock.patch('mlsriracha.plugins.azureml.train.Path.mkdir', autospec=True) as mock_mkdir:
            mock_mkdir.return_value = True
            plugin = AzureMlTrain()
        
        response = plugin.finish()
        self.assertTrue(response)

    def test_azureml_data(self):
         with mock.patch('mlsriracha.plugins.azureml.train.Path.mkdir', autospec=True) as mock_mkdir:
            with mock.patch('mlsriracha.plugins.azureml.train.os.environ.get', autospec=True) as mock_env:
                mock_mkdir.return_value = True
                mock_env.return_value = './static/data/data.csv'
                plugin = AzureMlTrain()
               
                response = plugin.input_as_dataframe(channel='training')
                # print(response.size)
                self.assertTrue(response.size == 15)

    def test_azureml_folder(self):
         with mock.patch('mlsriracha.plugins.azureml.train.Path.mkdir', autospec=True) as mock_mkdir:
                mock_mkdir.return_value = True
                plugin = AzureMlTrain()
               
                response = plugin.log_artifact('model.pkl')
                print(response)
                self.assertTrue(response.find('/outputs/model/model.pkl') != -1)

    # def test_gcpvertex_plugin(self):
    #     plugin = AwsSageMakerTrain()
    #     response = plugin.finish()
    #     self.assertTrue(response)


from setuptools import find_packages, setup
setup(
    name='mlsriracha',
    version='0.0.1',
    description='A project to abstract the boilerplate required to deploy jobs in MLOps systems',
    author='Alex Chung',
    author_email='alex@socialg.tech',
    url='https://github.com/awcchungster/mlsriracha',
    packages=find_packages(),
    install_requires=[
        'google-cloud-storage>=1.40.0'
    ],
)


from mlctlsriracha.plugins.azureml.predict import AzureMlPredict
from mlctlsriracha.plugins.gcpvertex.predict import GcpVertexPredict
from mlctlsriracha.plugins.awssagemaker.predict import AwsSageMakerPredict

class PredictAdapter:
    def __init__(self, provider: str):
        self.provider_name = provider.lower()

        print('This is a prediction job')
        try: provider
        except NameError: 
            raise RuntimeError(f'{str} is not a valid provider')
        
        if provider.lower() == 'azureml':
            print('Using Azure ML as a provider')
            self.provider_obj = AzureMlPredict()
        elif provider.lower() == 'gcpvertex':
            print('Using GCP Vertex as a provider')
            self.provider_obj = GcpVertexPredict()
        elif provider.lower() == 'awssagemaker':
            print('Using AWS SageMaker as a provider')
            self.provider_obj = AwsSageMakerPredict()
        else:
            raise RuntimeError(f'{str} is not a valid provider')

    def model_artifact(self, filename: str):
        return self.provider_obj.model_artifact(filename)

    def endpoint_metadata(self):
        return self.provider_obj.endpoint_metadata()


from mlctlsriracha.plugins.azureml.train import AzureMlTrain
from mlctlsriracha.plugins.gcpvertex.train import GcpVertexTrain
from mlctlsriracha.plugins.awssagemaker.train import AwsSageMakerTrain

class TrainingAdapter:
    def __init__(self,
        provider: str):
        self.provider_name = provider.lower()

        print('This is a training job')
        try: provider
        except NameError: 
            raise RuntimeError(f'{str} is not a valid provider')

        if provider.lower() == 'azureml':
            print('Using Azure ML as a provider')
            self.provider_obj = AzureMlTrain()
        elif provider.lower() == 'gcpvertex':
            print('Using GCP Vertex as a provider')
            self.provider_obj = GcpVertexTrain()
        elif provider.lower() == 'awssagemaker':
            print('Using AWS SageMaker as a provider')
            self.provider_obj = AwsSageMakerTrain()
        else:
            raise RuntimeError(f'{str} is not a valid provider')
        
            
    def input_as_dataframe(self, channel: str):
        return self.provider_obj.input_as_dataframe(channel=channel)

    def artifact_path(self, filename: str):
        return self.provider_obj.log_artifact(filename)

    def log_param(self, params):
        pass

    def log_artifact(self, object, type='model'):
        pass

    def log_metric(self, params):
        pass 

    def finish(self):
        return self.provider_obj.finish()